import contextlib
import functools
import logging

from typing import Callable, Type, Tuple, Union, Optional
from typeguard import typechecked

from ppa import ui

from .display import TaskSteps

logger = logging.getLogger(__name__)


class CallbackFailed(Exception):
    pass


class ParameterError(Exception):
    pass


@typechecked
def error_handler(
    steps: Union[TaskSteps, None] = None,
    ignore: Union[Tuple[Type[Exception]], Type[Exception], tuple] = (),
    message: Optional[str] = None,
    callback: Optional[Callable] = None,
    proceed_steps: Optional[bool] = False,
):
    """
    A decorator that provides error handling and reporting for task functions.

    In its most basic form (called with no arguments), `@error_handler` will ensure that any exception
    is logged and raised, and that the user is delivered an error message in the PPA task interface.

    You can further customise the behaviour of the decorator by supplying a combination of the following
    keyword arguments.

    :param steps: An optional :py:class:`display.TaskSteps` table. If supplied, this decorator will
        call `task_failed()` on the table if any Exception is raised.
    :param ignore: An optional Exception, or tuple of Exceptions. If the decorated function raises an
        ignored Exception, no special behaviour will be performed, and the function will finish cleanly. By
        default *no Exceptions are ignored*.
    :param message: An optional error message to display in the PPA task interface should the decorated
        function raise an `unignored` Exception. If this is not supplied, the Exception message itself will
        be displayed.
    :param callback: An optional function to execute if the decorated function raises an `unignored` Exception.
    :param proceed_steps: An optional boolean. Defaults to False. If set to True, the decorator will call `proceed()`
        on the supplied `steps` table if the decorated function does not raise an `unignored` Exception. To use
        this argument you **must** supply a `steps` table.

    :raises ParameterError: if the decorator is called with `proceed_steps=True` but no `steps` parameter
        is provided.

    .. seealso:: The :ref:`control_flow:The Error Handler Decorator` documentation section for **much** more
        usage information, and detailed examples.
    """
    if proceed_steps and not steps:
        raise ParameterError(
            "The error_handler decorator was called with proceed_steps=True but no steps table has been provided."
        )

    if not isinstance(ignore, tuple):
        ignore = (ignore,)

    def _wrapper(func: Callable):
        @functools.wraps(func)
        def _fn(*args, **kwargs):
            raised = False
            try:
                return func(*args, **kwargs)
            except Exception as e:
                function_name = func.__name__

                if ignore:
                    # If ignore is not None, and the raised exception matches any
                    # exceptions in the ignore tuple, log some information and pass.
                    if any([isinstance(e, ignored) for ignored in ignore]):  # type: ignore
                        logger.info(f"function {function_name} raised the following exception which will be ignored")
                        logger.exception(e)
                        # There will be no value to return from the decorated function
                        # as it will have failed, so we can return None here.
                        return None

                # The raised exception is either not ignored, or there are
                # no ignored exceptions defined - report and raise.
                logger.info(f"function {function_name} raised caught exception type {type(e)}")
                if steps:
                    steps.task_failed()
                ui.output_error(f"Operation Failed: {message or str(e)}")
                raised = True
                raise e

            finally:
                if callback and raised:
                    # Attempt to execute the supplied callback function and
                    # raise a custom exception if it fails. The callback may
                    # be doing important teardown work and should fail the
                    # decorator if it raises.
                    ui.output_progress("Executing hook on task failure...", percent=None, element_id="callback")
                    try:
                        logger.info("executing callback function")
                        callback()
                        ui.output_progress("Executed clean-up hook successfully", percent=100, element_id="callback")
                        logger.info("callback function executed successfully")
                    except Exception as e:
                        ui.output_progress("Failed to execute clean-up hook", percent=0, element_id="callback")
                        ui.output_error(
                            "Failed to execute hook on task failure. Please view the logs for more information."
                        )
                        logger.error("failed to execute callback function")
                        logger.exception(e)
                        raise CallbackFailed(
                            f"Failed to execute callback function triggered when {e} "
                            "raised a caught exception. Please view the logs for more information."
                        )
                elif all([not raised, proceed_steps]):
                    # If there are steps after this one, call proceed().
                    final_step_index = len(steps.steps) - 1
                    if steps.current_step_index < final_step_index:
                        steps.proceed()
                    # If this is the last step, call task_finished().
                    elif steps.current_step_index == final_step_index:
                        steps.task_finished()
                    # Otherwise something weird has happened.
                    else:
                        raise ParameterError(
                            f"Cannot finalize or proceed steps table. "
                            f"Current step index: {steps.current_step_index}. "
                            f"Final step index: {final_step_index}."
                        )

        return _fn

    return _wrapper


@contextlib.contextmanager
def exception_translate(exception_map: dict):
    """
    A context manager that catches a supplied exception, and raises a new exception.

    :param exception_map: A dictionary of mappings. Each `key` in the dictionary should be a type of exception, and
        each key's `value` should be another exception to raise if the `key` exception is raised. See the code example
        below for more information.

    :raises: Defined in `exception_map`. The new exception mapped to the original raised exception.

    Example - Catch any `IndexError` and raise the custom exception `NoRecordsFound` with a defined message:

    .. code-block:: python

        with exception_translate(
            {IndexError: NoRecordsFound("Could not find any records using the supplied details")}
        ):
            perform_operation(arguments)


    .. seealso:: The :ref:`control_flow:Exception Translation` documentation section for more information.
    """
    try:
        yield
    except Exception as raised_exception:
        for original, new in exception_map.items():
            if isinstance(raised_exception, original):
                raise new
        raise raised_exception
    finally:
        pass
