import smtplib
import logging
import datetime
import base64
from email.mime.multipart import MIMEMultipart
from email.mime.image import MIMEImage
from email.mime.text import MIMEText
from typing import Union, List, Tuple

import validators
from textile import textile

from ppa import ui
from ppa_tools.vendor_icons import PPA_FULL

logger = logging.getLogger(f"ppa-tools.{__name__}")


class InvalidEmailAddress(Exception):
    pass


class InvalidEmailFormat(Exception):
    pass


def _validate_email(email: str) -> str:
    if not validators.email(email):
        message: str = f"Address {email} is not a valid email address."
        ui.output_error(message)
        raise InvalidEmailAddress(message)
    return email


def _validate_addresses(from_address, to_addresses) -> Tuple[str, str]:
    if not isinstance(from_address, str):
        message = "From address must be supplied as string, " f"got {type(from_address)}."
        ui.output_error(message)
        raise InvalidEmailFormat(message)
    _validate_email(from_address)

    if type(to_addresses) not in {list, str}:
        message = "To addresses must be supplied as list or string, " f"got {type(to_addresses)}."
        raise InvalidEmailFormat(message)

    if isinstance(to_addresses, list):
        list(map(_validate_email, to_addresses))
        to_addresses = ", ".join(to_addresses)
    else:
        _validate_email(to_addresses)

    return (from_address, to_addresses)


def form_email(from_address: str, to_addresses: Union[List, str], subject: str, message_html: str):
    from_address, to_addresses = _validate_addresses(from_address, to_addresses)
    email = MIMEMultipart("related")
    email["Subject"] = subject
    email["From"] = from_address
    email["To"] = to_addresses
    html_message: MIMEText = MIMEText(message_html, "html")
    png_data: bytes = base64.b64decode(PPA_FULL["base_64"])
    image = MIMEImage(png_data, "png")
    image.add_header("Content-ID", "<ppa>")
    image.add_header("Content-Disposition", "inline", filename="ppa.png")
    email.attach(html_message)
    email.attach(image)
    return email


def create(
    from_address: str, to_addresses: Union[List, str], subject: str, title: str, message_body: str
) -> MIMEMultipart:
    html = f"""
        <html>
            <head>
                <p style="text-align:center;">
                    <img src="cid:ppa" width="175"/>
                </p>
                <p style="text-align:center;font-weight:bold;font-size:20;">
                    {title}
                </p>
            </head>
            <body style="text-align:left;font-size:16;">
                <br/>
                <p>{textile(message_body, html_type="html")}</p>
                <footer>
                    <br/>
                    <br/>
                    <p style="text-align:center;font-weight:lighter;font-size:10;">
                        Email generated and delivered by Osirium PPA at {datetime.datetime.utcnow().strftime(
        '%H:%M:%S %d/%m/%Y')}
                    </p>
                </footer>
            </body>

        </html>
    """
    return form_email(from_address, to_addresses, subject, html)


def create_approval(
    from_address: str, to_addresses: Union[List, str], requester: str, message_body: str, wait_hook_url: str
) -> MIMEMultipart:
    html = f"""
        <html>
            <head>
                <p style="text-align:center;">
                    <img src="cid:ppa" width="175"/>
                </p>
                <p style="text-align:center;font-weight:bold;font-size:20;">
                    You have a new PPA approval request from {requester}
                </p>
            </head>
            <body style="text-align:left;font-size:16;">
                <br/>
                <p>{textile(message_body, html_type="html")}</p>
                <p>To approve this request click <a href="{wait_hook_url}?approve=true">here</a>.</p>
                <p>To reject this request click <a href="{wait_hook_url}?approve=false">here</a>.</p>
                <footer>
                    <br/>
                    <br/>
                    <p style="text-align:center;font-weight:lighter;font-size:10;">
                        Approval email generated and delivered by Osirium PPA at {datetime.datetime.utcnow().strftime(
        '%H:%M:%S %d/%m/%Y')}
                    </p>
                </footer>
            </body>

        </html>
    """
    return form_email(from_address, to_addresses, "PPA Approval Request", html)


def send(
    server: str,
    port: int,
    login: str,
    password: str,
    email: MIMEMultipart,
    start_tls: bool = True,
    smtp_auth: bool = True,
):
    to_addresses = email["To"].split(",")  # type: ignore
    logger.info(f"sending email using server {server}:{port}")
    logger.info(f'sending from {email["From"]}')
    logger.info(f"sending to {to_addresses}")

    def _send():
        with smtplib.SMTP(server, port=port) as smtp_server:
            smtp_server.ehlo()
            if start_tls:
                smtp_server.starttls()
            if smtp_auth:
                smtp_server.login(login, password)
            smtp_server.sendmail(str(email["From"]), to_addresses, email.as_string())

    if not isinstance(email, MIMEMultipart):
        raise TypeError("Email parameter must be of type email.mime.multipart.MIMEMultipart.")
    _send()
