"""
CSV

This module contains actions used to create CSV spreadsheets.
"""

from typing import List, Dict, Any


def create_from_table(header: List[str], rows: List[List[str]]) -> str:
    r"""Generate CSV data from a table header & rows.

!!! question "Minimum Plugin Version: 5.0.0"

??? info "Input"

    - __header:__ a list of table column headers

    - __rows:__ a list of lists where each inner list contains each cell value

??? success "Output"

    The generated CSV as a string.

??? example

    ```YAML linenums="1"
    - ppa_tools.csv.create_from_table:
        header:
          - Given Name
          - Surname
          - Locked
        rows:
          - - John
            - Smith
            - True
          - - Bruce
            - Wayne
            - False
      save: users_csv

    - ppa_tools.files.write_text_file:
        name: users.csv
      load:
        contents: users_csv

    - ppa.events.send_email:
        subject: Users Report
        html: >
          <html>
            <head>Users Report</head>
            <body>The CSV spreadsheet is attached to this email.</body>
          </html>
        attachments:
          - users.csv
    ```

!!! tip "Header & Row Length"

    Each list inside the `rows` input must be the same length as the supplied `header`.
    """
    lines = [",".join(header)]
    list(map(lambda row: lines.append(",".join(row)), rows))
    return "\n".join(lines)


def create_from_dictionaries(header: List[str], keys: List[str], dictionaries: List[Dict[str, Any]]) -> str:
    r"""Generate CSV data from a list of dictionaries & keys.

!!! question "Minimum Plugin Version: 5.0.0"

??? info "Input"

    - __header:__ a list of table column headers

    - __keys:__ a list of corresponding dictionary keys, one for each `header`

    - __dictionaries:__ a list of dictionaries to get the values of each `key` from

??? success "Output"

    The generated CSV as a string.

??? example

    - Creating a CSV spreadsheet of Active Directory users in the Domain Admins group

    - Writing the CSV spreadsheet to a file

    - Sending the file as an email attachment

    ```YAML linenums="1"
    - active_directory.groups.get_users:
        distinguishedName: CN=Domain Admins,CN=Users,DC=Example,DC=Domain
      load:
        domain_controller: domain_controller
      save: domain_admins

    - ppa_tools.csv.create_from_dictionaries:
        header:
          - Common Name
          - SAM Account Name
          - Email Address
        keys:
          - cn
          - sAMAccountName
          - mail
      load:
        dictionaries: domain_admins
      save: domain_admins_csv

    - ppa_tools.files.write_text_file:
        name: domain_admins.csv
      load:
        contents: domain_admins_csv

    - ppa.events.send_email:
        subject: Domain Admins Report
        html: >
          <html>
            <head>PPA Domain Admins Report</head>
            <body>The CSV spreadsheet is attached to this email.</body>
          </html>
        attachments:
          - domain_admins.csv
    ```

!!! tip "Headers & Keys"

    The `header` & `keys` lists must be the same length, & the attributes must be in the same order.

    If a supplied key is missing from a dictionary, its value will be blank.
    """

    def _get_cell_value(d, k):
        return str(d.get(k)) if d.get(k) is not None else ""

    return "\n".join(
        [", ".join(header)]
        + list(map(lambda d: ", ".join([_get_cell_value(d, key).replace(",", "") for key in keys]), dictionaries))
    )
